package uk.ac.leeds.ccg.jianhuij.nn;
/*
 * Layer.java
 * @author  jin
 * @version 1.0
 */

public class Layer extends java.lang.Object implements java.io.Serializable {
    
    /**
     * For storing the number of neurons
     */
    int neuroNumber = 0;
    
    /**
     * For storing the number of inputs
     */
    int inputNumber = 0;
    
    /**
     * Learning rate
     */
    double rate = 0.25d;
    
    /**
     * Alfa value for sigmoid function
     */
    double alfa = 1.0d;
    
    /**
     * Arrays for storing weights
     */
    double[][] lastLayerOldWeight = null;
    double[][] thisLayerOldWeight = null;
    
    /**
     * Arrays for storing error signals
     */
    double[] lastLayerErrorSignal = null;
    double[] thisLayerErrorSignal = null;
    
    /**
     * Array that stores the neurons
     */
    Neuron[] neuros = null;
    
    /**
     * Array for storing neuron output values
     */
    double[] fire = null;
    
    /**
     * Array for storing neuron input values
     */
    double[] input = null;
    
    /**
     * Initialises Layer
     */
    private void ini( int neuroNumber, int inputNumber, double rate, double alfa ) {
        this.neuroNumber = neuroNumber;
        this.inputNumber = inputNumber;
        this.rate = rate;
        this.alfa = alfa;
        fire = new double[ neuroNumber ];
        neuros = new Neuron[ neuroNumber ];
        for ( int i = 0; i < neuroNumber; i ++ ) {
            neuros[i] = new Neuron( inputNumber, rate, alfa );
        }
        thisLayerOldWeight = new double[ inputNumber ][ neuroNumber ];
        thisLayerErrorSignal = new double[ neuroNumber ];
    }
    
    /** Creates new Layer */
    public Layer( int neuroNumber, int inputNumber, double rate, double alfa ) {
        ini( neuroNumber, inputNumber, rate, alfa );
    }
    
    /**
     * Processes data through the layer
     */
    public void process( double[] input ) {
        for ( int i = 0; i < neuros.length; i ++ ) {
            neuros[ i ].process( input );
            fire[ i ] = neuros[ i ].getFire();
        }
    }
    
    /**
     * Returns fire
     */
    public double[] getFire(){
        return this.fire;
    }
    
    /**
     * Adjusts weights
     */
    public void WAdjust( double[][] lastLayerOldWeight, double[] lastLayerErrorSignal ){
        for( int i = 0; i < neuroNumber; i ++ ) {
            neuros[i].WAdjust( lastLayerErrorSignal, lastLayerOldWeight[i]);
        }
    }
    
    /**
     * Returns the old weights
     */
    public double[][] getOldWeight(){
        double[][] temp = new double[neuros.length][];
        for ( int i = 0; i < neuros.length; i ++ ) {
            temp[i]=neuros[i].getOldWeight();
        }
        for ( int i = 0; i < inputNumber; i ++ ) {
            for ( int j = 0; j < neuros.length; j ++ ) {
                thisLayerOldWeight[i][j] = temp[j][i+1];
            }
        }
        return thisLayerOldWeight;
    }
    
    /**
     * Returns the error signal
     */
    public double[] getErrorSignal(){
        for ( int i = 0; i < neuros.length; i ++ ) {
            thisLayerErrorSignal[i] = neuros[i].getErrorSignal();
        }
        return thisLayerErrorSignal;
    }
    
    /**
     * Returns the neurons
     */
    Neuron[] getNeuro() {
        return neuros;
    }
    
    /**
     * Sets the neurons
     */
    void setNeuro( Neuron[] neuros ) {
        this.neuros = neuros;
    }
    
    /**
     * Returns the number of neurons in layer
     */
    public int getNeuroNumber() {
        return getNeuro().length;
    }
    
    /**
     * Returns alfa
     */
    public double getAlfaValue() {
        return alfa;
    }
    
    /**
     * Sets alfa
     */
    public void setAlfaValue( double alfa ) {
        this.alfa = alfa;
    }
    
    /**
     * Returns rate
     */
    public double getLearningRate() {
        return rate;
    }
    
    /**
     * Sets rate
     */
    protected void setLearningRate( double rate ) {
        this.rate = rate;
        Neuron[] neurons = getNeuro();
        if ( neurons == null ) { return; }
        else {
            for ( int i = 0; i < neurons.length; i ++) {
                neurons[i].setLearningRate( rate );
            }
        }
    }
    
    /**
     * Returns a single array with all weights
     */
    public double[] getWeight() {
        neuroNumber = getNeuroNumber();
        inputNumber = getInputNumber();
        double[] temp = new double[ neuroNumber * ( inputNumber + 1 ) ];
        for( int i = 0; i < neuroNumber; i ++ ){
            double[] tempWeight = neuros[i].getWeight();
            for( int j = 0; j < inputNumber + 1 ; j ++ ) {
                temp[ i * ( inputNumber + 1 ) + j ] = tempWeight[ j ];
            }
        }
        return temp;
    }
    
    /**
     * Set weights
     */
    public void setWeight( double[] weights ) throws Exception {
        int weightNumber = neuros[0].getWeight().length;
        if ( weights.length != weightNumber * getNeuroNumber() ) {
            throw new Exception("weights number does not match that of the old weights");
        }
        double[] tempWeight = new double[ weightNumber ];
        for( int i = 0; i < neuros.length; i ++ ) {
            for( int j = 0; j < weightNumber; j ++ ) {
                tempWeight[ j ] = weights[ j + ( weightNumber * i ) ];
            }
            neuros[ i ].setWeight( tempWeight );
        }
    }
    
    /**
     * Returns inputNumber
     */
    public int getInputNumber() {
        return this.inputNumber;
    }
    
    /**
     * Returns input
     */
    public double[] getInput() {
        return input;
    }
    
    /**
     * sets input
     */
    public void setInput( double[] input ) {
        this.input = input;
    }
}
