package uk.ac.leeds.ccg.jianhuij.nn;

import java.util.*;
import java.io.*;
/*
 * NeuroNetwork.java
 * @author  Jianhui Jin
 * @version 1.0
 */

public class NeuroNetwork extends java.lang.Object implements java.io.Serializable {

    /**
     * For storing the number of layers (hidden layers + input layer)
     */
    private int layerNumber = 0;
    /**
     * learning rate for backpropagation training and alfa value in the sigmoid
     */
    private double rate = 0.25,  alfa = 1;
    // input layer and hidden layers
    private Layer[] layers = null;
    // outputlayer using the same class as other layers do
    private Layer outputLayer = null;
    private int inputNumber = 0;
    private double noDataValue = -9999.0;
    /*
     * root mean square of the error between true training output and fire
     */
    private double rMSE = 1.0d;

    //OutNeuro outNeuron=null;
    public NeuroNetwork(Layer[] layers, Layer outputLayer, double rate, double alfa) {
        setLayers(layers);
        setLayerNumber(layers.length);
        setLearningRate(rate);
        setAlfaValue(alfa);
        setOutputLayer(outputLayer);
        setInputNumber(layers[ 0].getInputNumber());
    }

    public NeuroNetwork(int[] neuronNumber, int inputNumber, double rate, double alfa, int outputNumber) {
        Layer[] layers = new Layer[neuronNumber.length];
        layers[0] = new Layer(neuronNumber[0], inputNumber, rate, alfa);
        if (layers.length > 1) {
            for (int i = 1; i < layers.length; i++) {
                layers[i] = new Layer(neuronNumber[i], neuronNumber[i - 1], rate, alfa);
            }
        }
        Layer outputLayer = new Layer(outputNumber, neuronNumber[neuronNumber.length - 1], rate, alfa);
        setLayers(layers);
        setLayerNumber(layers.length);
        setOutputLayer(outputLayer);
        setLearningRate(rate);
        setAlfaValue(alfa);
        setInputNumber(inputNumber);
    }

    public NeuroNetwork(int[] neuronNumber, int inputNumber, double rate, double alfa) {
        Layer[] layers = new Layer[neuronNumber.length];
        layers[0] = new Layer(neuronNumber[0], inputNumber, rate, alfa);
        if (layers.length > 1) {
            for (int i = 1; i < layers.length; i++) {
                layers[i] = new Layer(neuronNumber[i], neuronNumber[i - 1], rate, alfa);
            }
        }
        Layer outputLayer = new Layer(1, neuronNumber[neuronNumber.length - 1], rate, alfa);
        setLayers(layers);
        setLayerNumber(layers.length);
        setAlfaValue(alfa);
        setOutputLayer(outputLayer);
        setInputNumber(inputNumber);
        setLearningRate(rate);
    }

    public NeuroNetwork(Object[] parameter) {
        int[][] neuronAndInputNumber = (int[][]) parameter[0];
        Object[] rateAndAlfa = (Object[]) parameter[1];
        double[] weights = (double[]) parameter[2];

        Layer[] layers = new Layer[neuronAndInputNumber.length - 1];
        Layer outLayer = null;

        for (int i = 0; i < layers.length; i++) {
            layers[i] = new Layer((int) neuronAndInputNumber[i][0], (int) neuronAndInputNumber[i][1], ((Double) rateAndAlfa[0]).doubleValue(), ((Double) rateAndAlfa[1]).doubleValue());
        }
        outLayer = new Layer((int) neuronAndInputNumber[layers.length][0], (int) neuronAndInputNumber[layers.length][1], ((Double) rateAndAlfa[0]).doubleValue(), ((Double) rateAndAlfa[1]).doubleValue());

        this.setLayers(layers);
        this.setLayerNumber(layers.length);

        this.setOutputLayer(outLayer);

        this.setLearningRate(((Double) rateAndAlfa[0]).doubleValue());
        this.setAlfaValue(((Double) rateAndAlfa[1]).doubleValue());
        try {
            this.setWeightByGeneticLearning(weights);
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }
        ;
        this.setInputNumber(layers[0].getInputNumber());
    }

    /*public NeuroNetwork(Layer[] layers, double rate, double alfa) {
    this.setLayers(layers);
    this.setLayerNumber(layers.length);
    this.setLearningRate(rate);
    this.setAlfaValue(alfa);
    //this.setOutputLayer(outputLayer);
    this.outNeuron=new OutNeuro(layers[layers.length-1].getNeuroNumber(),rate,alfa);
    }*/
    public void process(double[] x) {
        Layer[] layers = this.getLayer();
        Layer outLayer = this.getOutputLayer();
        //OutNeuro outNeuron=this.getOutputNeuron();
        layers[ 0].process(x);

        double[] temp = layers[ 0].getFire();

        for (int i = 1; i < layers.length; i++) {
            layers[i].process(temp);
            temp = layers[i].getFire();
        }

        outLayer.process(temp);

    }

    public double[] getFire(double[] x) {
        double[] fire = new double[getOutputLayer().getNeuroNumber()];
        process(x);
        fire = this.getOutputLayer().getFire();
        return fire;
    }

    private void WAdjust(double[] y, double[] fire) {
        Layer[] layers = getLayer();
        Layer outLayer = getOutputLayer();
        //OutNeuro outNeuron=this.getOutputNeuron();
        // construct the error for output layer since all the neuron and layers
        // in this architecture is the same whereas the backpropagation works differently
        // in the output layer and the other layers
        // in those layers except the output layer, the error signal is the summation of
        // the weighted error signals of the next layer.
        // but for output layer, the error signal directly from the training output y[].
        // in order to use the same neuron and layer, we have to construct two array.
        // one is the error signal array, the other is the weights, from the next layer.
        // but for each output neuron it only needs the corresponding training output.
        // so we will make a weights array[][], only one weight is 1.0, others are 0.0
        // for example, weight[0][0] = 1.0, weight[0][1] = 0.0, weight[0][2] = 0.0 and so on.
        // hence, each output neuron will only get its own training signal.
        double[] errorForOutLayer = new double[y.length];
        for (int i = 0; i < y.length; i++) {
            errorForOutLayer[i] = y[i] - fire[i];
        }
        // the initial value of a element in an array is 0.0
        // so we only need to set the one needed.
        double[][] lastWeightForOutLayer = new double[y.length][y.length];
        for (int i = 0; i < y.length; i++) {
            lastWeightForOutLayer[i][i] = 1.0d;
        }
        outLayer.WAdjust(lastWeightForOutLayer, errorForOutLayer);
        double[][] tempWeight = outLayer.getOldWeight();
        double[] tempErrorSignal = outLayer.getErrorSignal();
        for (int i = layerNumber - 1; i >= 0; i--) {
            layers[i].WAdjust(tempWeight, tempErrorSignal);
            tempWeight = layers[i].getOldWeight();
            tempErrorSignal = layers[i].getErrorSignal();
        }
    }

    public double[] training(double[] x, double[] y) {
        double[] fire = getFire(x);
        boolean nodata = false;
        double noDataValue = getNoDataValue();
        for (int i = 0; i < fire.length; i++) {
            if ((fire[i] == noDataValue) || (y[i] == noDataValue)) {
                nodata = true;
                break;
            }
        }

        if (!nodata) {
            WAdjust(y, fire);
        }

        return fire;
    }

    public Layer[] getLayer() {
        return this.layers;
    }

    public void setWeightByGeneticLearning(double[] weights) throws Exception {
        Layer[] layers = this.getLayer();
        Layer outLayer = this.getOutputLayer();
        //OutNeuro outNeuron=this.getOutputNeuron();
        int usedWeightNumber = 0;
        for (int i = 0; i < layerNumber; i++) {
            double[] tempWeight = new double[layers[i].getWeight().length];
            for (int j = 0; j < tempWeight.length; j++) {
                tempWeight[j] = weights[usedWeightNumber + j];
            }
            this.layers[i].setWeight(tempWeight);
            usedWeightNumber = usedWeightNumber + layers[i].getWeight().length;

        }

        double[] tempWeight = new double[weights.length - usedWeightNumber];
        for (int i = 0; i < tempWeight.length; i++) {
            tempWeight[i] = weights[usedWeightNumber + i];
        }
        outLayer.setWeight(tempWeight);

    }

    double[] getWeightForGeneticLearning() {
        Layer[] layers = getLayer();
        Layer outLayer = getOutputLayer();
        //OutNeuro outNeuron=this.getOutputNeuron();

        double[] temp = null;
        Vector tempVector = new Vector();

        for (int i = 0; i < layerNumber; i++) {
            double[] tempWeight = layers[i].getWeight();
            for (int j = 0; j < tempWeight.length; j++) {
                tempVector.add(new Double(tempWeight[j]));
            }
        }

        double[] tempWeight = outLayer.getWeight();
        for (int i = 0; i < tempWeight.length; i++) {
            tempVector.add(new Double(tempWeight[i]));
        }

        Object[] tempDouble = tempVector.toArray();
        temp = new double[tempDouble.length];
        for (int i = 0; i < temp.length; i++) {
            temp[i] = ((Double) tempDouble[i]).doubleValue();
        }

        return temp;

    }

    public static NeuroNetwork load(File nnParameterObjectFile) throws IOException, ClassNotFoundException {

        ObjectInputStream objectStream = new ObjectInputStream(new BufferedInputStream(new FileInputStream(nnParameterObjectFile)));

        Object nn = objectStream.readObject();

        objectStream.close();

        return new NeuroNetwork((Object[]) nn);

    }

    public static void save(NeuroNetwork nn, File out) throws IOException {
        ObjectOutputStream objectStream = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(out)));
        objectStream.writeObject(nn.getNNParameter());
        objectStream.close();

    }

    /** set all the parameter to the object[]
     * the first object is a int[layer][layer's parameter], int[0][0] is the first layer neuro number\n
     * int[0][1] is the input number , int[0][2] is the next layer neuro number
     * the second object include learning rate is Double object \n
     * alfa value is Double object \n
     * the above 2 object will be in a object[] which is the second object \n
     * weight[] will be the third object \n
     * the return is a object[3] \n
     *
     */
    public Object[] getNNParameter() {
        Layer[] layers = this.getLayer();
        Layer outLayer = this.getOutputLayer();
        //OutNeuro outNeuron=this.getOutputNeuron();

        int[][] object1 = new int[layers.length + 1][2];
        for (int i = 0; i < layers.length; i++) {
            object1[i][0] = layers[i].getNeuroNumber();
            object1[i][1] = layers[i].getInputNumber();

        }
        object1[object1.length - 1][0] = outLayer.getNeuroNumber();
        object1[object1.length - 1][1] = outLayer.getInputNumber();

        Object[] object2 = {new Double(getLearningRate()), new Double(getAlfaValue())};
        Object[] ob = new Object[3];
        ob[0] = object1;
        ob[1] = object2;
        ob[2] = this.getWeightForGeneticLearning();
        return ob;
    }

    public int getLayerNumber() {
        return this.getLayer().length;
    }

    public double getAlfaValue() {
        return this.alfa;
    }

    public double getLearningRate() {
        return this.rate;
    }

    protected void setLayerNumber(int layerNumber) {
        this.layerNumber = layerNumber;
    }

    protected void setLearningRate(double learningRate) {
        this.rate = learningRate;
        Layer[] hiddenLayers = getLayer();
        Layer outputLayer = getOutputLayer();
        if (hiddenLayers == null) {
            return;
        } else {
            for (int i = 0; i < hiddenLayers.length; i++) {
                hiddenLayers[i].setLearningRate(learningRate);
            }
        }
        outputLayer.setLearningRate(learningRate);
    }

    public void setAlfaValue(double alfa) {
        this.alfa = alfa;
    }

    public void setLayers(Layer[] layers) {
        this.layers = layers;
    }

    /** add a new neuron to the indexed layer
     */
    public void addNeuron(int index) {
    }

    public void setOutputLayer(Layer outputLayer) {
        this.outputLayer = outputLayer;

    }

    public Layer getOutputLayer() {
        return this.outputLayer;
    }

    public double[][] runNN(double[][] x) {
        double[][] fire = new double[x.length][];
        for (int i = 0; i < x.length; i++) {
            double[] temp = getFire(x[i]);
            fire[i] = new double[temp.length];
            for (int j = 0; j < temp.length; j++) {
                fire[i][j] = temp[j];
            }
        }
        return fire;
    }

    public double[][] trainNN(double[][] x, double[] y) throws Exception {
        double[][] newY = new double[y.length][1];

        for (int i = 0; i < y.length; i++) {
            newY[i][ 0] = y[i];
        }

        double[][] fire = trainNN(x, newY);
        setRMSE(NNUtility.getRMS(fire, newY));
        return fire;

    }

    public double[][] trainNN(double[][] x, double[][] y) throws Exception {
        double[][] fire = null;

        //double[] oldWeight = getWeightForGeneticLearning();

        //double oldRMSE = NNUtility.getRMS( runNN( x ), y );

        for (int i = 0; i < x.length; i++) {
            training(x[i], y[i]);
        }

        fire = runNN(x);

        double newRMSE = NNUtility.getRMS(fire, y);

        //setLearningRate( newRMSE );

        setRMSE(newRMSE);

        return fire;



    }

    private void setRMSE(double rmse) {
        this.rMSE = rmse;
    }

    public double getRMSE() {
        return rMSE;
    }

    public int getInputNumber() {
        return inputNumber;
    }

    public void setInputNumber(int inputNumber) {
        this.inputNumber = inputNumber;
    }

    public void setNoDataValue(double nodatavalue) {
        this.noDataValue = nodatavalue;

    }

    public double getNoDataValue() {
        return this.noDataValue;

    }

    /**
     * return a string look like "input=3&layernumber=2&layer0=3&layer1=5&outputnumber=2&rate=1.0&alfa=1.0&genetic=100&back=100"
     * it is a string used as query parameter in servlet request URL, but don't have the character of "?"
     **/
    public String toURLQueryParameterString(int geneticRound, int backRound) {
        Layer[] layers = getLayer();
        Layer outputLayer = getOutputLayer();
        int inputNumber = getInputNumber();
        String parameter = "input=" + inputNumber;
        parameter = parameter + "&" + "layernumber=" + layers.length;
        for (int i = 0; i < layers.length; i++) {
            parameter = parameter + "&" + "layer" + i + "=" + layers[i].getNeuroNumber();
        }
        parameter = parameter + "&" + "output=" + outputLayer.getNeuroNumber();
        parameter = parameter + "&" + "rate=" + getLearningRate();
        parameter = parameter + "&" + "alfa=" + getAlfaValue();
        parameter = parameter + "&" + "genetic=" + geneticRound;
        parameter = parameter + "&" + "back=" + backRound;
        return parameter;
    }
    /* public OutNeuro getOutputNeuron() {
    return this.outNeuron;
    }*/
}
