package uk.ac.leeds.ccg.jianhuij.nn;
/*
 * Neuron.java
 * @author  Jianhui Jin
 * @version 1.0
 */

public class Neuron extends java.lang.Object implements java.io.Serializable {

    /**
     * For storing learning rate
     */
    double rate = 0.25d;
    /**
     * For storing inputNumber
     */
    int inputNumber = 0;
    /**
     * For storing alfa value
     */
    double alfa = 1.0d;
    /**
     * For storing current weights
     */
    double[] weight = null;
    /**
     * For storing old weights
     */
    double[] oldWeight = null;
    /**
     * For storing dot product value
     */
    double dot = 0.0d;
    /**
     * For storing fire value
     */
    double fire = 0.0d;
    /**
     * For storing input
     */
    double[] input = null;
    /**
     * For storing error signal
     */
    double errorSignal = 0.0d;
    /**
     * For storing last time wieght adjustment for momentum
     */
    double[] lastTimeWeightChange = null;

    private void ini(int inputNumber, double rate, double alfa) {
        setInputNumber(inputNumber);
        setLearningRate(rate);
        setAlfaValue(alfa);
        // initialise the weights in neuron using random numbers
        double[] weight = new double[inputNumber + 1];
        double[] oldWeight = new double[inputNumber + 1];
        for (int i = 0; i < inputNumber + 1; i++) {
            weight[i] = Math.random();
            oldWeight[i] = 0.0d;
        }
        setWeight(weight);
        setOldWeight(oldWeight);
        lastTimeWeightChange = new double[inputNumber + 1];
    }

    /**
     * Creates new OutNeuro
     */
    public Neuron(int inputNumber, double rate, double alfa) {
        ini(inputNumber, rate, alfa);
    }

    /**
     * Process data
     */
    public void process(double[] input) {
        // store the values for backprobagation
        this.input = input;
        // dot product starts from weight[0] that is a constant.
        this.dot = this.weight[0];
        // weighting each input and set the sum of these weighted value as dot product
        for (int i = 0; i < inputNumber; i++) {
            this.dot = this.dot + this.weight[i + 1] * input[i];
        }
        // the fire out of a sigmoid function that is the transfer function,
        // using dot product as input.
        fire = 1.0d / (1.0d + StrictMath.exp(-1.0d * this.alfa * this.dot));
    }

    /**
     * Adjusts weights by back propogation
     */
    public void WAdjust(double[] lastLayerError, double[] lastLayerWeight) {
        double lastLayerErrorSum = 0.0d;
        double[] weight = getWeight();
        double[] oldWeight = getOldWeight();
        // double[] lastTimeWeightChange = this.getLastTimeWeightChange();
        // the total error caused to the next layer because of this neuron is the summation of
        // the weights of this neuron in the connected neurons in the next layer mutiply the error of the
        // corresponding neuron in the next layer
        // which means this neuron contributes part of the error in each next layer neuron by its weights.
        // for backpropagation, those errors need to be summated by their weights.
        for (int i = 0; i < lastLayerError.length; i++) {
            lastLayerErrorSum += lastLayerError[i] * lastLayerWeight[i];
        }
        double fire = getFire();
        double errorSignal = getAlfaValue() * lastLayerErrorSum * (1.0d - fire) * fire;
        setErrorSignal(errorSignal);
        double learningRate = getLearningRate();
        oldWeight[0] = weight[0];
        weight[0] = weight[0] + errorSignal * learningRate + lastTimeWeightChange[0] * 0.9;
        //weight[0] = weight[0] + errorSignal * learningRate;
        lastTimeWeightChange[0] = weight[0] - oldWeight[0];
        int inputNumber = getInputNumber();
        double[] input = getInput();
        for (int i = 0; i < inputNumber; i++) {
            oldWeight[i + 1] = weight[i + 1];
            weight[i + 1] = weight[i + 1] + errorSignal * learningRate * input[i] + lastTimeWeightChange[i + 1] * 0.9;
            weight[i + 1] = weight[i + 1] + errorSignal * learningRate * input[i];
            lastTimeWeightChange[i + 1] = weight[i + 1] - oldWeight[i + 1];
        }
        setWeight(weight);
        setOldWeight(oldWeight);
        setLastTimeWeightChange(lastTimeWeightChange);
    }

    /**
     * Returns fire
     */
    public double getFire() {
        return fire;

    }

    /**
     * Returns old weight
     */
    public double[] getOldWeight() {
        double[] temp = new double[oldWeight.length];
        for (int i = 0; i < oldWeight.length; i++) {
            temp[i] = oldWeight[i];
        }
        return temp;
    }

    /**
     * Returns error signal
     */
    public double getErrorSignal() {
        return this.errorSignal;
    }

    /**
     * Returns weights
     */
    public double[] getWeight() {
        double[] temp = new double[weight.length];
        for (int i = 0; i < weight.length; i++) {
            temp[i] = this.weight[i];
        }
        return temp;
    }

    /**
     * Sets weights
     */
    public void setWeight(double[] weights) {
        this.weight = new double[getInputNumber() + 1];
        for (int i = 0; i < weights.length; i++) {
            this.weight[i] = weights[i];
        }
    }

    /**
     * Sets fire
     */
    private void setFire(double fire) {
        this.fire = fire;
    }

    /**
     * Returns rate
     */
    private double getLearningRate() {
        return this.rate;
    }

    /**
     * Sets rate
     */
    protected void setLearningRate(double rate) {
        this.rate = rate;
    }

    /**
     * Sets inputNumber
     */
    private void setInputNumber(int inputNumber) {
        this.inputNumber = inputNumber;
    }

    /**
     * Returns inputNumber
     */
    private int getInputNumber() {
        return this.inputNumber;
    }

    /**
     * Sets alfa
     */
    private void setAlfaValue(double alfa) {
        this.alfa = alfa;
    }

    /**
     * Returns alfa
     */
    private double getAlfaValue() {
        return this.alfa;
    }

    /**
     * Sets dot
     */
    private void setDotProduct(double dot) {
        this.dot = dot;
    }

    /**
     * Returns dot
     */
    private double getDotProduct() {
        return this.dot;
    }

    /**
     * Returns input
     */
    public double[] getInput() {
        return input;
    }

    /**
     * Sets input
     */
    private void setInput(double[] input) {
        this.input = input;
    }

    /**
     * Sets errorSignal
     */
    private void setErrorSignal(double errorSignal) {
        this.errorSignal = errorSignal;
    }

    /**
     * Returns lastTimeWeightChange
     */
    private double[] getLastTimeWeightChange() {
        double[] temp = new double[lastTimeWeightChange.length];
        for (int i = 0; i < lastTimeWeightChange.length; i++) {
            temp[i] = lastTimeWeightChange[i];
        }
        return temp;
    }

    /**
     * Sets lastTimeWeightChange
     */
    private void setLastTimeWeightChange(double[] weightChange) {
        if (weightChange != null) {
            lastTimeWeightChange = new double[weightChange.length];
            for (int i = 0; i < weightChange.length; i++) {
                lastTimeWeightChange[i] = weightChange[i];
            }
        }
    }

    /**
     * Sets oldWeight
     */
    public void setOldWeight(double[] oldWeight) {
        this.oldWeight = new double[oldWeight.length];
        for (int i = 0; i < oldWeight.length; i++) {
            this.oldWeight[i] = oldWeight[i];
        }
    }
}
