/*
 * SimulatedAnnealing.java
 *
 * Created on 16 September 2003, 21:13
 */
package uk.ac.leeds.ccg.jianhuij.nn;

import java.util.Random;

/**
 * DOCUMENT ME!
 *
 * @author jianhuij
 */
public class SimulatedAnnealing {

    /**
     * Creates a new instance of SimulatedAnnealing
     *
     * @param startTemp DOCUMENT ME!
     * @param endTemp DOCUMENT ME!
     * @param tempReducePer DOCUMENT ME!
     * @param startIteration DOCUMENT ME!
     * @param restart DOCUMENT ME!
     * @param initial DOCUMENT ME!
     */
    double startTemp = 100;
    double endTemp = 0.001;
    double tempNow = startTemp;
    double tempReducePer = 0.1;
    double startIteration = 100;
    int restart = 5;
    NeuroNetwork nn = null;
    //training data x and y
    double[][] x = null;
    double[][] y = null;
    // randomized data x and y
    double[][] randomX = null;
    double[][] randomY = null;
    
    double[] weight = null;
    double[] bestWeight = null;
    double bestErrorForOneRestart = Double.MAX_VALUE;
    double bestErrorForAllRestart = Double.MAX_VALUE;
    Random random = new Random();

    public SimulatedAnnealing(double startTemp, double endTemp,
            double tempReducePer, double startIteration, int restart,
            NeuroNetwork initial, double[][] x, double[][] y) {
        this.startIteration = startIteration;
        this.startTemp = startTemp;
        this.endTemp = endTemp;
        this.tempReducePer = tempReducePer;
        this.weight = initial.getWeightForGeneticLearning();
        this.bestWeight = initial.getWeightForGeneticLearning();
        this.nn = initial;
        this.x = x;
        this.y = y;

    }
    
    /**
     * 
     *  randomize input and output, basically the order of the input and output 
     * by swapping
     * 
     */
    private void randomizeInputAndOutput() {
        for ( int i = 0; i < x.length; i ++ ) {
            int swapIndex1 = random.nextInt(x.length);
            double[] tempX = x[swapIndex1];
            double[] tempY = y[swapIndex1];
            
            int swapIndex2 = random.nextInt(x.length);
            x[swapIndex1] = x[swapIndex2];
            y[swapIndex1] = y[swapIndex2];
            
            x[swapIndex2] = tempX;
            y[swapIndex2] = tempY;
        }
    }

    public NeuroNetwork run() {
        
        randomizeInputAndOutput();
        
        
        int i = 0;
        //for ( int i = 0; i < restart; i ++ ) {
        double maxTemp = this.startTemp + (this.startTemp * restart * 1.0d) / 10;
        int maxIteration = (int) (this.startIteration + (this.startIteration * restart * 1.0d) / 10);
        bestErrorForOneRestart = Double.MAX_VALUE;
        for (tempNow = maxTemp; tempNow > endTemp;
                tempNow = tempNow - (tempReducePer * tempNow)) {
            temperatureDropOneStep(maxTemp, maxIteration);
        }
        if (bestErrorForAllRestart > bestErrorForOneRestart) {
            bestErrorForAllRestart = bestErrorForOneRestart;
            for (int l = 0; l < weight.length; l++) {
                bestWeight[l] = weight[l];
            }
            try {
                nn.setWeightByGeneticLearning(bestWeight);
            } catch (Exception e) {
                System.out.println("Set Weight in SA temperature drop once" + e);
            }
            i--;
            if (i < 0) {
                i = 0;
            }
        }
        System.out.println("Restart " + i);
        //}

//        double[][] fire = nn.runNN(x);
//        for (int j = 0; j < fire.length; j++) {
//            System.out.println("Fire" + fire[j][0] ); //+ " Target " + y[j][0]);
//        }
        return nn;
    }

    private NeuroNetwork temperatureDropOneStep(double maxTemp, int maxIteration) {
        double newError = Double.MAX_VALUE;
        double oldError = Double.MAX_VALUE;
        double CT = (tempNow * 0.01) / maxTemp;

        for (int i = 0;
                i < ((maxIteration * (maxTemp - (tempNow / 2))) / (maxTemp -
                endTemp)); i++) {
            // the frequency of the change of the household in the BHPSHousehold
            // determined by the temperature. the higher temperature, the more times
            // household change in the Vector
            double[] oldWeight = new double[weight.length];
            for (int l = 0; l < oldWeight.length; l++) {
                oldWeight[l] = weight[l];
                if (random.nextDouble() > 0.5) {
                    weight[l] = weight[l] + random.nextDouble() * tempNow * 1.0d / startTemp;
                } else {
                    weight[l] = weight[l] - random.nextDouble() * tempNow * 1.0d / startTemp;
                }
            }
            try {
                nn.setWeightByGeneticLearning(weight);
            } catch (Exception e) {
                System.out.println("Set Weight in SA temperature drop once" + e);
            }
            double[][] fire = nn.runNN(x);

            newError = getError(fire, y);

            double accept = Math.exp((double) -(newError - oldError) / CT);

            if (((newError - oldError) <= 0) || (random.nextDouble() < accept)) { // if new error smaller than actual error, or if trial of worst error approved
                oldError = newError;

                if (oldError < bestErrorForOneRestart) {
                    // it is usually profitable to continue iterating at a given temperature as long as the function is still decreasing.(Timothy Masters)
                    i--;
                    bestErrorForOneRestart = oldError;
                }
            } else {
                // deferent way, so just assign the old Vector back, since we changed more than one
                // household in the new way
                weight = oldWeight;
                try {
                    nn.setWeightByGeneticLearning(weight);
                } catch (Exception e) {
                    System.out.println("Set Weight in SA temperature drop once" + e);
                }
            }
        }
        System.out.println("NewError: " + newError + " Best " + bestErrorForOneRestart + " Temperature at Now:" + tempNow);
        return nn;
    }

    private double getError(double[][] fire, double[][] target) {
     //   return getRMS(fire, target) + getAverageRelativeError(fire, target);
           return getAverageTABError(fire, target);
    }

    private double getAverageTABError(double[][] fire, double[][] target) {
        double averageTABError = 0.0;

        for (int i = 0; i < fire[0].length; i++) {
            double errorSquareSum = 0.0;
            for (int j = 0; j < fire.length; j++) {
                errorSquareSum = errorSquareSum + Math.abs(fire[j][i] - target[j][i]);
            }
            averageTABError = averageTABError + errorSquareSum / fire.length;
        }
        return averageTABError / fire[0].length;
    }

    private double getAverageRelativeError(double[][] fire, double[][] target) {
        double RMSRelative = 0.0;

        for (int i = 0; i < fire[0].length; i++) {

            double errorSquareSum = 0.0;

            for (int j = 0; j < fire.length; j++) {
                errorSquareSum = errorSquareSum + Math.abs((fire[j][i] - target[j][i]) / target[j][i]);
            }

            RMSRelative = RMSRelative + errorSquareSum / fire.length;

        }

        return RMSRelative / fire[0].length;
    }

    private double getRMS(double[][] fire, double[][] target) {
        double RMSSum = 0.0;

        for (int i = 0; i < fire[0].length; i++) {

            double errorSquareSum = 0.0;

            for (int j = 0; j < fire.length; j++) {
                errorSquareSum = errorSquareSum + (fire[j][i] - target[j][i]) * (fire[j][i] - target[j][i]);
            }

            RMSSum = RMSSum + StrictMath.sqrt(errorSquareSum / (fire.length - 1));

        }

        return RMSSum / fire[0].length;

    }
}
